/*always included to use the MPI library*/
#include "mpi.h"

/*this includes the definition of a token*/
#include "message.h"

/**********************************************************************/
/*                                                                    */
/* Ring of Two Processes -                                            */
/*     a Simple MPI Demonstration Program                             */
/* by John Langford  Thurs. 20 July, 1995                             */
/* written to accomplish the same activities as a Maisie examply by   */
/*                                                                    */
/*             Mani Chandy/John Thornley,                             */
/*             Computer Science Dept., Caltech.                       */
/*                                                                    */
/* Simulates two processes in a ring.  A token is passed around the   */
/* ring NUM_ITERATIONS times.  The processes are initiated with       */
/* identification numbers 0 and 1.  Process 0 holds the token         */
/* initially and finally.                                             */
/*                                                                    */
/**********************************************************************/

void Fill_token(MESSAGE *token);
void Print_token(MESSAGE *token);
void Make_type(MPI_Datatype *token);

/* the number of times that you go around the ring*/
#define NUMITS 12

/*main is started on all of the nodes*/

void main(int argc,char **argv)
{
  MESSAGE token;
  MPI_Datatype mpi_token;
  int rank,i;
  MPI_Status status;
/*this is always called before running MPI functions*/
  MPI_Init( &argc,&argv);
/*Another piece of set up*/
  MPI_Comm_rank(MPI_COMM_WORLD,&rank);
/*Another initialization function which allows you
to pass and recieve tokens of arbitrary size and shape*/
  Make_type(&mpi_token);
/*sets the token shape and size to be passed*/
  MPI_Type_commit(&mpi_token);
  if (rank==0)/* Node 0 is handled a little differently than other nodes */
    {
      Fill_token(&token);
      for (i=0; (i<NUMITS); i++)
/*MPI_Sendrecv_replace sends a token then waits to recieve it*/
	MPI_Sendrecv_replace(&token,1,mpi_token,1,RANDOM_TAG,1,RANDOM_TAG,MPI_COMM_WORLD,&status);
/*arguments are Send arguments, then Recv arguments with all duplicated 
arguments deleted*/
      Print_token(&token);
    }
  else
    {
      MPI_Recv(&token,1,mpi_token,0,RANDOM_TAG,MPI_COMM_WORLD,&status);
      for (i=0; (i<(NUMITS-1)); i++)
	MPI_Sendrecv_replace(&token,1,mpi_token,0,RANDOM_TAG,0,RANDOM_TAG,MPI_COMM_WORLD,&status);
      MPI_Send(&token,1,mpi_token,0,RANDOM_TAG,MPI_COMM_WORLD);
    }
  MPI_Finalize;
}

void Fill_token(MESSAGE *token)
{
  sprintf(token->string,"I am full.");
}

void Print_token(MESSAGE *token)
{
  printf("%s\n",token->string);
}

void Make_type(MPI_Datatype *token)
{
  MESSAGE msg[2];
  int blocklengths[NUMFIELDS+1]={20,1,30,1};
  MPI_Aint disps[NUMFIELDS+1];
  MPI_Datatype types[NUMFIELDS+1]= {MPI_CHAR,MPI_INT,MPI_DOUBLE,MPI_UB};
  int i,base;
  MPI_Address(msg[0].string,disps);
  MPI_Address(&msg[0].strlen,disps+1);
  MPI_Address(msg[0].random_stuff,disps+2);
  MPI_Address(msg+1,disps+3);
  base = disps[0];
  for (i=0; i<NUMFIELDS + 1; i++) {
    disps[i] = disps[i]-base;
  }
  MPI_Type_struct(NUMFIELDS+1,blocklengths,disps,types,token);
}
